/*
 * Crude test driver for processing the VST and MCT testvector files
 * generated by the CMVP RNGVS product.
 *
 * Note the input files are assumed to have a _very_ specific format
 * as described in the NIST document "The Random Number Generator
 * Validation System (RNGVS)", May 25, 2004.
 *
 */
#include <openssl/opensslconf.h>

#include <openssl/bn.h>
#include <openssl/fips.h>
#include <openssl/err.h>
#include <openssl/modes.h>
#include <string.h>
#include <ctype.h>

#include "fips_utl.h"

void die(char *mes)
{
    fprintf(stderr, mes);
    exit(1);
}

void process(char *req, char *rsp)
{
    char buf[2048], lbuf[2048];
    unsigned char result[2048];
    unsigned char *K = NULL;
    unsigned char *P = NULL;
    unsigned char *C = NULL;
    unsigned plaintext_len, ciphertext_len;
    unsigned key_len;
    char *end;
    AES_KEY aes_key;
    char *keyword, *value;
    long l;
    int length;
    int inverse = 0;
    block128_f f;

    FILE *in = fopen(req, "r");
    FILE *out = fopen(rsp, "w");

    if (!in || !out) {
        die("Can't open input or output file\n");
    }

    while(fgets(buf, sizeof(buf), in) != NULL)
    {
	fputs(buf,out);

	if (!parse_line(&keyword, &value, lbuf, buf)) {
            /* might be a header, check if inverse cipher function is requested */
	    if(strstr(buf, "inverse")) {
                inverse = 1;
            }
            continue;
        }

	if(!strcmp(keyword, "[PLAINTEXT LENGTH"))
        {
            end = value + strlen(value) - 1;
            /* remove trailing ] */
            if (*end == ']')
                *end = 0;
	    plaintext_len = atoi(value) / 8;
            ciphertext_len = plaintext_len + 8;
        }
        /* key */
	else if(!strcmp(keyword, "K"))
        {
	    K = hex2bin_m(value, &l);
            key_len = strlen(value) / 2;
        }
        /* plaintext */
	else if(!strcmp(keyword, "P"))
        {
            /* Wrap, we have a key and a plaintext */
	    P = hex2bin_m(value, &l);
            if (inverse) {
                if (AES_set_decrypt_key(K, key_len*8, &aes_key))
                    die("Can't set AES decrypt key.\n");
                f = (block128_f)AES_decrypt;
            } else {
                if (AES_set_encrypt_key(K, key_len*8, &aes_key))
                    die("Can't set AES encrypt key.\n");
                f = (block128_f)AES_encrypt;
            }
            length = CRYPTO_128_wrap(&aes_key, NULL, result, P, plaintext_len, f);
            if (!length)
                die("Wrapping failed.\n");
            OutputValue("C", result, length, out, 0);
        }
        /* ciphertext */
	else if(!strcmp(keyword, "C"))
	{
            /* Unwrap, we have a key and a ciphertext */
	    C = hex2bin_m(value, &l);
            if (inverse) {
                if (AES_set_encrypt_key(K, key_len*8, &aes_key))
                    die("Can't set AES encrypt key.\n");
                f = (block128_f)AES_encrypt;
            } else {
                if (AES_set_decrypt_key(K, key_len*8, &aes_key))
                    die("Can't set AES decrypt key.\n");
                f = (block128_f)AES_decrypt;
            }
            length = CRYPTO_128_unwrap(&aes_key, NULL, result, C, ciphertext_len, f);
            if (!length) {
                fprintf(out, "FAIL" RESP_EOL);
            } else {
                OutputValue("P", result, length, out, 0);
            }
        }
    }
}

int main(int argc,char **argv)
{
    if(argc != 3)
    {
	fprintf(stderr,"%s Req Rsp\n",argv[0]);
	exit(1);
    }
    if(!FIPS_mode_set(1))
    {
	do_print_errors();
	exit(1);
    }

    process(argv[1], argv[2]);

    return 0;
}
